@@ -1,4 +1,17 @@
 #!/usr/bin/env python3
+"""
+pystrict3: a Python3 code plausibility checker
+"""
+
+import ast
+import sys
+import builtins
+import logging
+import operator
+
+from .funcchecker import FuncLister, list_documented_parameters, max_documented_returns, CallLister, ModuleCallLister
+from .classchecker import ClassPropertiesLister
+
 """
 BSD 2-Clause License
 
@@ -28,33 +41,18 @@ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 """
 
-import ast
-import sys
-import builtins
-import logging
-import operator
-
-from .funcchecker import FuncLister, list_documented_parameters, max_documented_returns, CallLister, ModuleCallLister
-from .classchecker import ClassPropertiesLister
 first_item_getter = operator.itemgetter(0)
 
 preknown = set(builtins.__dict__).union({'__doc__', '__file__', '__name__', '__annotations__', '__dict__', '__builtins__'})
 
-"""
-def flat_walk(node):
-    yield node
-    if not isinstance(node, ast.FunctionDef) and not isinstance(node, ast.AsyncFunctionDef) and not isinstance(node, ast.ClassDef) and not isinstance(node, ast.Lambda) \
-            and not isinstance(node, ast.Import) and not isinstance(node, ast.ImportFrom):
-
-        for child in ast.iter_child_nodes(node):
-            yield from flat_walk(child)
-"""
 
 def get_arg_ids(node):
+    """List variable names mentioned in a node."""
     if hasattr(node, 'args') and isinstance(node.args, ast.arguments):
         args = node.args
         for arg in args.args:
             yield arg.arg
+            del arg
         if hasattr(args, 'kwonlyargs'):
             for arg in args.kwonlyargs:
                 yield arg.arg
@@ -63,7 +61,9 @@ def get_arg_ids(node):
         if getattr(args, 'kwarg', None) is not None:
             yield args.kwarg.arg
 
+
 def get_assigned_ids(node):
+    """List variable names being assigned in a node."""
     if isinstance(node, ast.Name) and not isinstance(node.ctx, ast.Del):
         yield node.id
     if hasattr(node, 'target'):
@@ -81,7 +81,9 @@ def get_assigned_ids(node):
         for el in node.elts:
             yield from get_assigned_ids(el)
 
+
 def get_deleted_ids(node):
+    """List variable names being deleted in a node."""
     if isinstance(node, ast.Name) and isinstance(node.ctx, ast.Del):
         yield node.id
     if hasattr(node, 'target'):
@@ -90,12 +92,14 @@ def get_deleted_ids(node):
         for target in node.targets:
             yield from get_deleted_ids(target)
 
+
 def is_container(node):
+    """Check if node contains code."""
     return isinstance(node, ast.FunctionDef) or isinstance(node, ast.AsyncFunctionDef) or isinstance(node, ast.ClassDef) or isinstance(node, ast.Lambda) or isinstance(node, ast.Expr) or isinstance(node, ast.Call)
 
 
 def get_all_ids(node):
-    """get all ids defined in names, tuples and targets here
+    """Get all ids defined in names, tuples and targets here.
 
     :param node: AST node
     """
@@ -122,6 +126,7 @@ def get_all_ids(node):
             args = node.args
             for arg in args.args:
                 yield arg.arg
+                del arg
             if hasattr(args, 'kwonlyargs'):
                 for arg in args.kwonlyargs:
                     yield arg.arg
@@ -132,10 +137,13 @@ def get_all_ids(node):
         else:
             for arg in node.args:
                 yield from get_all_ids(arg)
+                del arg
     if hasattr(node, 'value') and not isinstance(node, ast.Attribute):
         yield from get_all_ids(node.value)
 
+
 def block_terminates(statements):
+    """Check if code block stops the execution."""
     for statement in statements:
         if isinstance(statement, ast.Return):
             return True
@@ -147,6 +155,7 @@ def block_terminates(statements):
 
 class FuncDocVerifier(ast.NodeVisitor):
     """Compares parameter names in function definition with docstring."""
+
     def __init__(self, filename):
         self.filename = filename
         self.log = logging.getLogger('pystrict3.funcdoc')
@@ -160,14 +169,16 @@ class FuncDocVerifier(ast.NodeVisitor):
             # look for methods:
             if isinstance(child, ast.FunctionDef):
                 self.class_methods.append(child)
-    
+        self.generic_visit(node)
+
     def visit_FunctionDef(self, node):
         arguments = node.args
-        module_docstring = ast.get_docstring(node)
-        if module_docstring is None:
+        func_docstring = ast.get_docstring(node, clean=True)
+        if func_docstring is None:
+            self.log.debug('%s:%d:no docstring for function "%s"', self.filename, node.lineno, node.name)
             return
-        
-        documented_parameters = list_documented_parameters(module_docstring)
+
+        documented_parameters = list_documented_parameters(func_docstring)
         function_arguments = [arg.arg for arg in arguments.args]
         if node in self.class_methods:
             arguments = function_arguments[1:]
@@ -187,39 +198,54 @@ class FuncDocVerifier(ast.NodeVisitor):
                     self.filename, node.lineno, arg, node.name))
                 self.undocumented_parameters_found |= True
 
-        all_returns = [(return_tuple_length, return_node) for return_tuple_length, return_node in self.walk_tree(node.body) if return_tuple_length > 1]
+        all_returns = [
+            (return_tuple_length, return_node)
+            for return_tuple_length, return_node in self._walk_tree(node.body)
+            if return_tuple_length > 1]
         if len(all_returns) == 0:
             return
         return_tuple_length, return_node = max(all_returns, key=first_item_getter)
         if return_tuple_length > 1:
             # see if a return tuple is documented
-            num_documented_returns = max_documented_returns(module_docstring) or 1
-            self.log.debug("%s returns at most %s values; %d documented", node.name, return_tuple_length, num_documented_returns)
-            if return_tuple_length > num_documented_returns:
+            num_documented_returns = max_documented_returns(func_docstring)
+            self.log.debug("%s returns at most %s values; %s documented", node.name, return_tuple_length, num_documented_returns)
+            if num_documented_returns is not None and return_tuple_length > num_documented_returns:
                 names = [getattr(el, 'id', 'var') for el in return_node.elts]
-                sys.stderr.write('%s:%d: ERROR: function "%s" does not document return of %d elements as in line %d: (%s)\n' % (
-                    self.filename, node.lineno, node.name, return_tuple_length, return_node.lineno, ', '.join(names)))
-                self.undocumented_returns_found |= True
+                if num_documented_returns == 1:
+                    sys.stderr.write('%s:%d: WARNING: function "%s" may not document return of %d elements as in line %d: (%s)\n' % (
+                        self.filename, node.lineno, node.name, return_tuple_length, return_node.lineno, ', '.join(names)))
+                else:
+                    sys.stderr.write('%s:%d: ERROR: function "%s" does not document return of %d elements as in line %d: (%s)\n' % (
+                        self.filename, node.lineno, node.name, return_tuple_length, return_node.lineno, ', '.join(names)))
+                    self.undocumented_returns_found |= True
+        self.generic_visit(node)
 
-
-    def walk_tree(self, nodes):
+    def _walk_tree(self, nodes):
         for node in nodes:
             if isinstance(node, ast.Return):
-                #print("  inside:", node, node.value, len(getattr(node.value, 'elts', [])))
                 yield len(getattr(node.value, 'elts', [])), node.value
 
             if hasattr(node, 'body'):
                 body = node.body if isinstance(node.body, list) else [node.body]
-                yield from self.walk_tree(body)
+                yield from self._walk_tree(body)
 
             if getattr(node, 'orelse', []) != []:
-                yield from self.walk_tree(node.orelse)
+                yield from self._walk_tree(node.orelse)
 
             if getattr(node, 'finalbody', []) != []:
-                yield from self.walk_tree(node.finalbody)
+                yield from self._walk_tree(node.finalbody)
 
 
 class NameAssignVerifier():
+    """Verify name definitions and access.
+
+    Makes sure variables that are defined have not been defined before,
+    and variables being deleted have been defined before.
+
+    :param filename: file name
+    :param allow_variable_reuse: allow redefinition of variables
+    """
+
     def __init__(self, filename, allow_variable_reuse=False):
         self.filename = filename
         self.log = logging.getLogger('pystrict3.nameassign')
@@ -231,17 +257,18 @@ class NameAssignVerifier():
         self.known_checked = 0
 
     def variable_unknown_found(self, lineno, name):
+        """Report a undefined variable `name` was found in line `lineno`."""
         sys.stderr.write('%s:%d: ERROR: Variable unknown: "%s"\n' % (self.filename, lineno, name))
         self.found_variable_unknown |= True
         raise Exception(name)
 
     def assert_unknown(self, name, known, lineno, override_with_builtins=False):
-        """unified error message verifying `name` is in set `known`"""
+        """Report error if `name` is already in set `known`."""
         self.log.debug("assert_unknown: %s, %s", name, lineno)
         assert name is not None
         # the nested if allows deleting builtins and re-defining them afterwards
         if name in known:
-            previous_state, previous_lineno = known[name]
+            previous_state, _ = known[name]
             if name in preknown:
                 if not override_with_builtins:
                     sys.stderr.write('%s:%d: ERROR: Overwriting builtin variable: "%s"\n' % (
@@ -250,7 +277,7 @@ class NameAssignVerifier():
             elif name == '_':
                 # throwaway variable is allowed to be overridden
                 pass
-            elif previous_state is not None and previous_state == True:
+            elif previous_state is not None and previous_state:
                 self.found_variable_reused |= True
                 if self.allow_variable_reuse:
                     sys.stderr.write('%s:%d: WARNING: Variable "%s" set previously in line %d may have changed meaning\n' % (
@@ -262,8 +289,7 @@ class NameAssignVerifier():
             self.unknown_checked += 1
 
     def check_new_identifiers(self, elements, node, known):
-        """add newly defined variables and verify that the accessed ones are defined in known"""
-        #old_known = dict(**known)
+        """Add newly defined variables and verify that the accessed ones are defined in known"""
         add = {}
         forget = set()
         self.log.debug('check_new_identifiers: known: %s', known.keys() - preknown)
@@ -280,7 +306,8 @@ class NameAssignVerifier():
                     elif hasattr(name, 'name'):
                         name = name.name
                     self.log.debug('check_new_identifiers: name %s', name)
-                    self.assert_unknown(name, known, el.lineno,
+                    self.assert_unknown(
+                        name, known, el.lineno,
                         override_with_builtins=getattr(node, 'module', '') == 'builtins')
                     name = name.split('.')[0]
                     add_here[name] = el.lineno
@@ -292,7 +319,6 @@ class NameAssignVerifier():
                     for name in get_assigned_ids(target):
                         if isinstance(getattr(target, 'ctx'), ast.Del):
                             self.log.debug('check_new_identifiers: del target name %s', name)
-                            #print("-%s" % name)
                             if name in add:
                                 del add[name]
                             if name in add_here:
@@ -368,20 +394,20 @@ class NameAssignVerifier():
 
         for name, lineno in add.items():
             if name not in known:
-                #print('+%s' % name)
+                # print('+%s' % name)
                 known[name] = lineno
         for name in forget:
             if name in known:
-                #print('-%s' % name)
+                # print('-%s' % name)
                 del known[name]
         self.log.debug('known: %s', known.keys() - preknown)
 
-
     def walk_tree(self, nodes, preknown={}, depth=0):
         known_nodes = dict(**preknown)
+        depthstr = ' ' * depth
         terminated = False
         for node in nodes:
-            #print(); import pprintast; pprintast.pprintast(node)
+            # print(); import pprintast; pprintast.pprintast(node)
             if isinstance(node, ast.Return):
                 terminated |= True
                 break
@@ -389,25 +415,27 @@ class NameAssignVerifier():
                 if node.value.func.value.id == 'sys' and node.value.func.attr == 'exit':
                     terminated |= True
                     break
-            self.log.debug('%swalk tree: %s[args:%s, assigned:%s] (terminating: %d)', '  '*depth, type(node).__name__, set(get_arg_ids(node)), set(get_assigned_ids(node)), terminated)
+            self.log.debug(
+                '%swalk tree: %s[args:%s, assigned:%s] (terminating: %d)', depthstr,
+                type(node).__name__, set(get_arg_ids(node)), set(get_assigned_ids(node)), terminated)
 
             # modified = getattr(node, 'decorator_list', []) != []
-            
+
             if not is_container(node):
                 self.check_new_identifiers([node], node, dict(**known_nodes))
             elif hasattr(node, 'name'):
-                self.log.debug('%s+node name: %s', '  '*depth, node.name)
+                self.log.debug('%s+node name: %s', depthstr, node.name)
                 self.assert_unknown(node.name, known_nodes, node.lineno)
                 known_nodes[node.name] = (True, node.lineno)
 
             if not isinstance(node, ast.AugAssign) and not isinstance(node, ast.Name):
                 for name in get_assigned_ids(node):
-                    self.log.debug('%s+node assigned name: %s', '  '*depth, name)
+                    self.log.debug('%s+node assigned name: %s', depthstr, name)
                     if isinstance(node, ast.Assign):
                         self.assert_unknown(name, known_nodes, node.lineno)
                     known_nodes[name] = (True, node.lineno)
             for name in get_deleted_ids(node):
-                self.log.debug('%s+node deleted name: %s', '  '*depth, name)
+                self.log.debug('%s+node deleted name: %s', depthstr, name)
                 known_nodes[name] = (False, node.lineno)
 
             if hasattr(node, 'generators'):
@@ -415,10 +443,10 @@ class NameAssignVerifier():
                 known_nodes2 = dict(**known_nodes)
                 for g in node.generators[::-1]:
                     for name in get_assigned_ids(g):
-                        self.log.debug('%s+generator %s name: %s', '  '*depth, g, name)
+                        self.log.debug('%s+generator %s name: %s', depthstr, g, name)
                         self.assert_unknown(name, known_nodes2, node.lineno)
                         known_nodes2[name] = (True, node.lineno)
-                    self.walk_tree([g.target], known_nodes2, depth+1)
+                    self.walk_tree([g.target], known_nodes2, depth + 1)
             if hasattr(node, 'test'):
                 # ignore result, because variables do not leak out
                 self.check_new_identifiers([node.test], node, dict(**known_nodes))
@@ -427,13 +455,13 @@ class NameAssignVerifier():
                 self.check_new_identifiers([node.iter], node, dict(**known_nodes))
             if hasattr(node, 'value'):
                 # ignore result, because variables do not leak out
-                self.walk_tree([node.value], dict(**known_nodes), depth+1)
+                self.walk_tree([node.value], dict(**known_nodes), depth + 1)
                 # self.check_new_identifiers([node.value], node, dict(**known_nodes))
-            #if hasattr(node, 'target'):
-            #    for name in get_assigned_ids(node.target):
-            #        self.log.debug('+node target %s', name)
-            #        self.assert_unknown(name, known_nodes, node.lineno)
-            #        known_nodes[name] = (True, node.lineno)
+            # if hasattr(node, 'target'):
+            #     for name in get_assigned_ids(node.target):
+            #         self.log.debug('+node target %s', name)
+            #         self.assert_unknown(name, known_nodes, node.lineno)
+            #         known_nodes[name] = (True, node.lineno)
 
             nodes_to_add = {}
             nbranches = 0
@@ -448,21 +476,22 @@ class NameAssignVerifier():
             else:
                 for arg in getattr(node, 'args', []) + getattr(node, 'keywords', []):
                     for name in get_all_ids(arg):
-                        if name not in known_nodes or known_nodes[name][0] == False:
+                        # state may be unclear, in which case we should not perform the check
+                        if name not in known_nodes or known_nodes[name][0] not in (True, None):
                             self.variable_unknown_found(node.lineno, name)
                         known_nodes2[name] = (True, node.lineno)
             if hasattr(node, 'body'):
                 body = node.body if isinstance(node.body, list) else [node.body]
-                members = self.walk_tree(body, known_nodes2, depth+1)
+                members = self.walk_tree(body, known_nodes2, depth + 1)
                 if not is_container(node) and not block_terminates(body):
                     nbranches += 1
                     for name, var in members.items():
                         if name not in known_nodes or known_nodes[name][0] != var[0]:
                             nodes_to_add[name] = (nodes_to_add.get(name, (0, None))[0] + 1, var)
-                self.log.debug('%s+nodes from body: %s', '  '*depth, nodes_to_add)
+                self.log.debug('%s+nodes from body: %s', depthstr, nodes_to_add)
 
             for handler in getattr(node, 'handlers', []):
-                members = self.walk_tree(handler.body, known_nodes, depth+1)
+                members = self.walk_tree(handler.body, known_nodes, depth + 1)
                 if not block_terminates(handler.body):
                     nbranches += 1
                     known_nodes2 = dict(**known_nodes)
@@ -472,45 +501,45 @@ class NameAssignVerifier():
                     for name, var in members.items():
                         if name not in known_nodes or known_nodes[name][0] != var[0]:
                             nodes_to_add[name] = (nodes_to_add.get(name, (0, None))[0] + 1, var)
-                self.log.debug('%s+nodes from handlers: %s', '  '*depth, nodes_to_add)
+                self.log.debug('%s+nodes from handlers: %s', depthstr, nodes_to_add)
 
             if getattr(node, 'orelse', None) is not None:
-                self.log.debug('%s+nodes from else: %s', '  '*depth, node.orelse)
-                members = self.walk_tree([node.orelse], known_nodes, depth+1)
+                self.log.debug('%s+nodes from else: %s', depthstr, node.orelse)
+                members = self.walk_tree([node.orelse], known_nodes, depth + 1)
                 if not block_terminates([node.orelse]):
                     nbranches += 1
                     for name, var in members.items():
                         if name not in known_nodes or known_nodes[name][0] != var[0]:
                             nodes_to_add[name] = (nodes_to_add.get(name, (0, None))[0] + 1, var)
-                self.log.debug('%s+nodes from else: %s', '  '*depth, nodes_to_add)
+                self.log.debug('%s+nodes from else: %s', depthstr, nodes_to_add)
 
             for name, (nbranches_adding, var) in nodes_to_add.items():
                 if nbranches_adding == nbranches:
                     self.assert_unknown(name, known_nodes, node.lineno)
                     if not is_container(node):
-                        self.log.debug('%s+%s (from all branches)', '  '*depth, name)
+                        self.log.debug('%s+%s (from all branches)', depthstr, name)
                         known_nodes[name] = True, var[1]
                 else:
                     if not is_container(node):
                         known_nodes[name] = None, var[1]
-                        self.log.debug('%s+%s (in some branches)', '  '*depth, name)
+                        self.log.debug('%s+%s (in some branches)', depthstr, name)
 
             if getattr(node, 'finalbody', []) != []:
-                members = self.walk_tree(node.finalbody, known_nodes, depth+1)
+                members = self.walk_tree(node.finalbody, known_nodes, depth + 1)
                 if not block_terminates(node.finalbody):
                     for name, var in members.items():
                         if name not in known_nodes or known_nodes[name][0] != var[0]:
-                            self.log.debug('%s+%s (from final)', '  '*depth, name)
+                            self.log.debug('%s+%s (from final)', depthstr, name)
                             known_nodes[name] = var
-                self.log.debug('%s+nodes from final: %s', '  '*depth, members)
-            
-            self.log.debug('%scurrent knowledge: %s', '  '*depth, known_nodes.keys() - preknown.keys())
+                self.log.debug('%s+nodes from final: %s', depthstr, members)
+
+            self.log.debug('%scurrent knowledge: %s', depthstr, known_nodes.keys() - preknown.keys())
 
         return known_nodes
 
 
 def main(filenames, module_load_policy='none', allow_variable_reuse=False):
-    """ Verify python files listed in filenames
+    """Verify python files listed in filenames.
 
     :param filenames: files to analyse
     :param module_load_policy: passed to ModuleCallLister
@@ -533,11 +562,11 @@ def main(filenames, module_load_policy='none', allow_variable_reuse=False):
         mcl = ModuleCallLister(filename=filename, load_policy=module_load_policy)
         mcl.visit(a)
         total_checked_calls += mcl.checked_calls
-        
+
         funcs = FuncLister(filename=filename)
         funcs.visit(a)
         known_functions.update(funcs.known_functions)
-        
+
         asts.append((filename, a))
         del filename, a
 
@@ -548,7 +577,7 @@ def main(filenames, module_load_policy='none', allow_variable_reuse=False):
         cl = CallLister(filename=filename, known_functions=known_functions)
         cl.visit(a)
         total_checked_calls += cl.checked_calls
-        
+
         print("%s: checking variables ..." % filename)
         nameassigner = NameAssignVerifier(filename, allow_variable_reuse=allow_variable_reuse)
         nameassigner.walk_tree(a.body, known)
